#!/usr/bin/env python3
# Licensed under the The AGPLv3 License (AGPLv3)
# Copyright (c) 2023 Forrest M. Hilton <forrestmhilton@gmail.com>
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from manim_lamination_builder.custom_json import custom_dump, read_file_to_laminations
from typing import List
from manim import (
    WHITE,
    Scene,
    Group,
    tempconfig,
)
import sys
import os

from manim.utils.file_ops import config

from manim_lamination_builder.lamination import Lamination

def group(laminations: List[Lamination]):
    group = Group(*[lamination.build() for lamination in laminations])
    group = group.arrange_in_grid()
    group.scale(
        1
        / max(group.width / config.frame_width + 0.01, group.height / config.frame_height + 0.01)
    )
    return group

class Main(Scene):
    def __init__(self, laminations: List[Lamination]):
        self.laminations = laminations
        super().__init__()

    def construct(self):
        self.camera.background_color = WHITE
        self.add(group(self.laminations))


if __name__ == "__main__":
    file = sys.argv[-1]
    if len(sys.argv) == 1:
        path = "/home/forrest/Desktop/manim_lamination_builder/test.json5"
    else:
        path = os.path.join(os.getcwd(), file)
    laminations = read_file_to_laminations(path)
    for lamination in laminations:
        lamination.auto_populate()

    with tempconfig({"quality": "medium_quality", "preview": True}):
        scene = Main(laminations)
        scene.render()



# Licensed under the The AGPLv3 License (AGPLv3)
# Copyright (c) 2023 Forrest M. Hilton <forrestmhilton@gmail.com>
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from manim import BLACK, TAU, Mobject, VMobject, tuplify

from manim_lamination_builder.points import NaryFraction, UnitPoint
from math import pi, tan


class Chord:
    min: NaryFraction
    max: NaryFraction

    def __init__(self, a: NaryFraction, b: NaryFraction):  # TODO: min max
        if a.to_float() < b.to_float():
            self.min = a
            self.max = b
        else:
            self.min = b
            self.max = a

    def crosses(self, other: "Chord") -> bool:
        if (
            self.max.to_float() <= other.min.to_float()
            or self.min.to_float() >= other.max.to_float()
        ):
            return False
        if (
            self.min.to_float() > other.min.to_float()
            and self.max.to_float() > other.max.to_float()
        ) or (
            self.min.to_float() < other.min.to_float()
            and self.max.to_float() < other.max.to_float()
        ):
            return True
        return False

    def handle_length(self) -> float:
        theta1 = self.min.to_angle()
        theta2 = self.max.to_angle()
        return handle_length(theta1, theta2)

    def build(self) -> Mobject:
        # https://pomax.github.io/bezierinfo/#circles_cubic
        ret = VMobject(color=BLACK)
        make_and_append_bezier(ret, self.min, self.max)
        return ret

    def __eq__(self, other: "Chord"):
        if isinstance(other, Chord):
            return self.min == other.min and self.max == other.max
        return False


def make_and_append_bezier(vmob: VMobject, theta1: UnitPoint, theta2: UnitPoint):
    """Add a cubic Bezier curve to a VMobject using given angles or NaryFractions."""

    a = theta1.to_cartesian()
    b = theta2.to_cartesian()

    handle_len = handle_length(theta1.to_angle(), theta2.to_angle())
    vmob.add_cubic_bezier_curve(a, a * (1 - handle_len), b * (1 - handle_len), b)


def handle_length(theta1: float, theta2: float) -> float:
    theta1, theta2 = theta1 % TAU, theta2 % TAU
    if theta1 > theta2:
        theta1, theta2 = theta2, theta1
    if abs(theta1 - theta2) - pi < 1e-8:
        # 2.2250738585072014e-308 is the smallest positive number
        theta1 += 1e-7
    # wraping edge case:
    if theta2 - theta1 > pi:
        theta1, theta2 = theta1 + pi, theta2 + pi
    delta_angle = pi - abs(theta1 - theta2)
    alpha = (theta1 + theta2) / 2
    r = tan(alpha - theta1)
    k = 4 / 3 * tan(delta_angle / 4)
    return r * k


a, b, c, d = tuplify(NaryFraction.from_string(4, "0_0").pre_images())

assert Chord(a, c).crosses(Chord(b, d))
assert not Chord(a, b).crosses(Chord(c, d))
assert Chord(b, d).crosses(Chord(a, c))
assert not Chord(c, d).crosses(Chord(a, b))
assert NaryFraction.from_string(4, "0_0") == NaryFraction.from_string(4, "0_0")
assert Chord(NaryFraction.from_string(4, "0_0"), b) in [
    Chord(NaryFraction.from_string(4, "0_0"), b)
]



"""
This is a re-implementation of lamination-builder that runs in python without dependencies on a browser and focusing on convenient generation of figures. Instead it uses Manim, which has several dependencies, which you are responsible for installing in accordance with Manim's installation instructions.
"""
from manim_lamination_builder.lamination import Lamination
from manim_lamination_builder.custom_json import (
    custom_dump,
    custom_parse,
    read_file_to_laminations,
    parse_lamination,
)
from manim_lamination_builder.points import FloatWrapper, UnitPoint, NaryFraction, sigma
from manim_lamination_builder.animation import AnimateLamination
from manim_lamination_builder.chord import Chord
from manim_lamination_builder.generate import (
    curried_colorize_with_respect_to,
    generate_sibling_portraits,
    generate_unicritical_lamination,
    remove_non_original_pollygons,
)
from manim_lamination_builder.main import group, Main



from manim import (
    ORIGIN,
    WHITE,
    Animation,
    Scene,
    tempconfig,
    Arc,
    Dot,
    VMobject,
    Mobject,
    TAU,
)
from manim_lamination_builder.chord import make_and_append_bezier
from manim_lamination_builder.custom_json import custom_dump, custom_parse
from manim_lamination_builder.lamination import Lamination
from manim_lamination_builder.points import FloatWrapper, angle_to_cartesian
from typing import Union


def lerp(a: float, b: float, alpha: float) -> float:
    "Linearly interpolate between two floats a and b using time value alpha."
    return (1 - alpha) * a + alpha * b


class AnimateLamination(Animation):
    def __init__(
        self,
        initial: Lamination,
        final: Lamination,
        start_mobject: Union[Mobject, None] = None,
        **kwargs,
    ) -> None:
        super().__init__(start_mobject or initial.build(), **kwargs)
        self.initial = initial
        self.final = final
        # TODO: check if laminations are compatible in terms of same length of properys

    def interpolate(self, alpha: float) -> None:
        circle = self.mobject.submobjects[0]
        assert isinstance(circle, Arc)
        center = circle.get_arc_center()
        radius = circle.radius

        count_arcs = 0
        count_polygons = 0
        for i, submobject in enumerate(self.mobject.submobjects):
            if isinstance(submobject, Dot):
                # Set the position of the Dot based on its position in the list of submobjects
                index_in_list_of_points = i - count_polygons - count_arcs
                point_initial = self.initial.points[index_in_list_of_points]
                point_final = self.final.points[index_in_list_of_points]
                submobject.arc_center = angle_to_cartesian(
                    lerp(point_initial.to_angle(), point_final.to_angle(), alpha)
                )
                submobject.generate_points()

            elif isinstance(submobject, Arc):
                count_arcs += 1
                continue
            else:
                assert isinstance(submobject, VMobject)
                # Set the position of all the points for the submobject based on its position in the list of submobjects
                count_polygons += 1
                index_in_list_of_vmobjects = i - count_arcs
                if index_in_list_of_vmobjects < len(self.initial.polygons):
                    polygon_initial = self.initial.polygons[index_in_list_of_vmobjects]
                    polygon_final = self.final.polygons[index_in_list_of_vmobjects]

                    submobject.reset_points()
                    for i in range(len(polygon_final)):
                        a = lerp(
                            polygon_initial[i].to_float(),
                            polygon_final[i].to_float(),
                            alpha,
                        )
                        b = lerp(
                            polygon_initial[(i + 1) % len(polygon_final)].to_float(),
                            polygon_final[(i + 1) % len(polygon_final)].to_float(),
                            alpha,
                        )
                        make_and_append_bezier(
                            submobject, FloatWrapper(a), FloatWrapper(b)
                        )
                else:  # occlusion
                    occlusion_initial = self.initial.occlusion
                    assert occlusion_initial is not None
                    # TODO: be more clever about wrap around
                    midpoint = (
                        occlusion_initial[0].to_float()
                        + occlusion_initial[1].to_float()
                    ) / 2
                    a = lerp(occlusion_initial[0].to_float(), midpoint, alpha)
                    b = lerp(occlusion_initial[1].to_float(), midpoint, alpha)
                    submobject.reset_points()
                    make_and_append_bezier(submobject, FloatWrapper(a), FloatWrapper(b))
                    circle.start_angle = b * TAU
                    circle.angle = ((a - b) % 1) * TAU
                    if circle.angle == 0:
                        circle.angle = TAU
                    circle.generate_points()

        for submobject in self.mobject.submobjects:
            if submobject is circle:
                continue
            submobject.points *= radius
            submobject.points += center
